﻿
//
// Copyright (c) 2014, Intel Corporation.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights 
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
// copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in	
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

// Authors:
//   Mickey Shkatov <mickey.shkatov@intel.com>

using System;
using System.Threading;
using System.IO;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Linq;
using System.Text;
using LibUsbDotNet;
using LibUsbDotNet.Main;
using LibUsbDotNet.LudnMonoLibUsb;
using NLog;
using Peach.Core;
using Peach;
using Peach.Core.Publishers;
using Peach.Core.IO;

namespace Peach.Core.Publishers
{
    [Publisher("UsbControlPublisher", true)]
	[Description("Preform control transfers, specifically 'get' and 'set' style transfers ")]
    [Parameter("VendorID", typeof(int), "The vendor ID of the usb device to connect to, Pit file input is a DECIMAL value")]
    [Parameter("ProductID", typeof(int), "The product ID of the usb device to connect to, Pit file input is a DECIMAL value")]
    [Parameter("Configuration", typeof(int), "the DECIMAL number for usb device configuration")]
    [Parameter("Interface", typeof(int), "the DECIMAL number for the usb device interface")]
    [Parameter("Read_requestType", typeof(int), "the DECIMAL number used for the read control transfer action", "192")] //0xC0
    [Parameter("Read_request", typeof(int), "the DECIMAL number used for read control transfer action", "0")]
    [Parameter("Read_value", typeof(int), "the DECIMAL number used for read control transfer action", "0")]
    [Parameter("Read_index", typeof(int), "the DECIMAL number used for read control transfer action", "0")]
    [Parameter("Read_length", typeof(short), "the DECIMAL number used for read control transfer action", "0")]

    [Parameter("Write_requestType", typeof(int), "the DECIMAL number used for the write control transfer action", "64")] //0x40
    [Parameter("Write_request", typeof(int), "the DECIMAL number used for the write control transfer action", "0")]
    [Parameter("Write_value", typeof(int), "the DECIMAL number used for the write control transfer action", "0")]
    [Parameter("Write_index", typeof(int), "the DECIMAL number used for the write control transfer action", "0")]
    [Parameter("Write_length", typeof(short), "the DECIMAL number used for the write control transfer action", "0")]

    /*
     * Using LibUsbDotNet http://sourceforge.net/projects/libusbdotnet/
     * 
     */
    public class UsbControlPublisher : Publisher
    {

        #region Constructor
        public UsbControlPublisher(Dictionary<string, Variant> args)
            : base(args)
        {

        } 
        #endregion

        #region Variable declarations 
        
        public UsbDevice MyUsbDevice;
        
        private static NLog.Logger logger = LogManager.GetCurrentClassLogger();
        protected override NLog.Logger Logger { get { return logger; } }
        private Stream _Stream;
        private string _Result;
        public int VendorID { get; set; }
        public int ProductID { get; set; }
        public int Configuration { get; set; }
        public int Interface { get; set; }
        public int ControlEndPoint { get; set; }

        public int Write_requestType { get; set; }
        public int Write_request { get; set; }
        public int Write_value { get; set; }
        public int Write_index { get; set; }
        public short Write_length { get; set; }
        public int Read_requestType { get; set; }
        public int Read_request { get; set; }
        public int Read_value { get; set; }
        public int Read_index { get; set; }
        public short Read_length { get; set; }

        public long SelfLength = 0;

        private UsbDeviceFinder MyUsbFinder = null;
        private ErrorCode ec = ErrorCode.None;
        #endregion Variable declarations
        
        #region Required get and set variables for Peach
        public override long Length
        {
            get
            {
                try
                {
                    return this._Stream.Length;
                }
                catch (Exception e)
                {
                    logger.Warn("Stream has issues: " + e.Message);
                    return 0;
                }


            }
        }
        public override long Position
        {
            get
            {
                try
                {
                    return this._Stream.Position;
                }
                catch (Exception e)
                {
                    logger.Warn("Stream has issues: " + e.Message);
                    return 0;
                }
            }
            set
            {
                this._Stream.Position = value;
            }
        }

        public override string Result
        {
            get
            {
                return _Result;
            }
            set
            {
                this._Result = value;
            }
        }
        #endregion Required get and set variables for Peach

        #region implemented overidden methods 

        protected override void OnStart()
        {
            try
            {
                MyUsbFinder = new UsbDeviceFinder(VendorID, ProductID);
                MyUsbDevice = UsbDevice.OpenUsbDevice(MyUsbFinder);
                //Try for a while to get the device, just in case it is slow or something, reduce dependency in "plugging it in" timing
                if (MyUsbDevice == null)
                    for (int i = 0; i < 20; i++)
                    {
                        Thread.Sleep(200);
                        MyUsbFinder = new UsbDeviceFinder(VendorID, ProductID);
                        MyUsbDevice = UsbDevice.OpenUsbDevice(MyUsbFinder);
                        if (MyUsbDevice != null) break;
                    }
                if (MyUsbDevice == null)
                    //No point to continue if the device is not found after we tried so hard to find it
                    throw new PeachException(string.Format("USB Publisher: Device Not Found. Vendor: {0} Product: {1} Interface: {2}",
                        VendorID, ProductID, Interface));

                IUsbDevice wholeUsbDevice = MyUsbDevice as IUsbDevice;
                if (!ReferenceEquals(wholeUsbDevice, null))
                {
                    wholeUsbDevice.SetConfiguration(Convert.ToByte(Configuration));
                    wholeUsbDevice.ClaimInterface(Interface);
                }

            }
            catch (Exception ex)
            {
                throw new PeachException(string.Format("USB Publisher: Device Not Found." + ex.Message));
            }
        }

        protected override void OnStop()
        {
            if (MyUsbDevice != null)
            {
                if (MyUsbDevice.IsOpen)
                {
                    IUsbDevice wholeUsbDevice = MyUsbDevice as IUsbDevice;
                    if (!ReferenceEquals(wholeUsbDevice, null))
                    {
                        wholeUsbDevice.ReleaseInterface(Interface);
                    }

                    MyUsbDevice.Close();
                }
                MyUsbDevice = null;
                UsbDevice.Exit();
            }
            base.OnStop();

        }

        protected override void OnInput()
        {
            try
            {
                
                byte[] readBuffer = new byte[512];
                byte[] ret = sendCtrlTransfer(readBuffer, (byte)Read_requestType, (byte)Read_request, Read_value, Read_index, Read_length, readBuffer.Length);
                if (ret != null)
                {
                    ret = ret.ToArray();
                    var str = ByteArrayToString(ret);

                    logger.Debug(str);

                    if (ret.Length != 0) 
                    {
                        logger.Debug("RECIEVED " + ret.Length + " BYTES FROM USB!");
                        logger.Debug("HEX - USB Publisher DEBUG  --> " + BitConverter.ToString(ret, 0, ret.Length).Replace("-", " "));
                        logger.Debug("ASCII - USB Publisher DEBUG  --> " + System.Text.Encoding.Default.GetString(ret, 0, ret.Length));
                    }
                    else logger.Debug("Nothing Read, recieved 0 Bytes from read operation");

                    this.Result = System.Text.Encoding.Default.GetString(ret, 0, ret.Length);
                    this._Stream = new MemoryStream(ret, 0, ret.Length, true, true);
                    this._Stream.Position = 0;
                    this.Position = 0;
                }
            }
            catch (Exception ex)
            {
                logger.Warn((ec != ErrorCode.None ? ec + ":" : String.Empty) + ex.Message);
            }
        }


        protected override void OnOutput(byte[] buffer, int offset, int count)
        {
            try
            {
                //We need to send count bytes from buffer starting at offset
                byte[] bytesToSend = new byte[count];
                for (int i = 0; i < count; i++)
                {
                    bytesToSend[i] = buffer[offset + i];
                }
                byte[] ret = sendCtrlTransfer(bytesToSend,(byte) Write_requestType,(byte) Write_request, Write_value, Write_index, Write_length, bytesToSend.Length);

                ret = ret.ToArray();
                var str = ByteArrayToString(ret);
                logger.Debug(str);
            }
            catch (Exception ex)
            {
                logger.Warn((ec != ErrorCode.None ? ec + ":" : String.Empty) + ex.Message);
            }
        }

        private byte[] sendCtrlTransfer(byte[] buffer, byte _requestType, byte _request, int _value, int _index, short _length, int _bufferLength)
        {
            ErrorCode ec = ErrorCode.None;

            try
            {
                
                int transferred =0 ;

                UsbSetupPacket setup = new UsbSetupPacket(_requestType, _request, _value, _index, _length);
                MyUsbDevice.ControlTransfer(ref setup, buffer, _bufferLength, out transferred);
                return buffer;


            }
            catch (Exception ex)
            {

                logger.Warn((ec != ErrorCode.None ? ec + ":" : String.Empty) + ex.Message);
                return null;
            }
         
        }

    
        #endregion implemented overidden methods

        #region Helper functions
        public static string ByteArrayToString(byte[] ba)
        {
            StringBuilder hex = new StringBuilder(ba.Length * 2);
            foreach (byte b in ba)
                hex.AppendFormat("{0:x2}", b);
            return hex.ToString();
        }
        #endregion Helper functions

        #region "Must have" overridden methods for the publisher

        public override void Flush()
        {
            this._Stream.Flush();
        }

        public override void SetLength(long value)
        {
            this._Stream.SetLength(value);
        }

        public override bool CanRead
        {
            get
            {
                return this._Stream.CanRead;
            }
        }

        public override bool CanWrite
        {
            get
            {
                return this._Stream.CanWrite;
            }
        }

        public override bool CanSeek
        {
            get
            {
                return this._Stream.CanSeek;
            }
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            this._Stream.Write(buffer, offset, count);
        }

        public override int ReadByte()
        {
            return this._Stream.ReadByte();
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            return this._Stream.Read(buffer, offset, count);
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            return this._Stream.Seek(offset, origin);
        }

        #endregion "Must have" overridden methods for the publisher

    }
}

// end
