﻿
//
// Copyright (c) 2014, Intel Corporation.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights 
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
// copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in	
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

// Authors:
//   Mickey Shkatov <mickey.shkatov@intel.com>

using System;
using System.Threading;
using System.IO;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Linq;
using System.Text;
using LibUsbDotNet;
using LibUsbDotNet.Main;
using LibUsbDotNet.LudnMonoLibUsb;
using NLog;
using Peach.Core;
using Peach;
using Peach.Core.Publishers;
using Peach.Core.IO;

namespace Peach.Core.Publishers
{
    [Publisher("UsbDataPublisher", true)]
	[Description("Send and recieve data on USB endpoints")]
    [Parameter("VendorID", typeof(int), "The vendor ID of the usb device to connect to, Pit file input is a DECIMAL value")]
    [Parameter("ProductID", typeof(int), "The product ID of the usb device to connect to, Pit file input is a DECIMAL value")]
    [Parameter("Configuration", typeof(int), "the DECIMAL number for usb device configuration")]
    [Parameter("Interface", typeof(int), "the DECIMAL number for the usb device interface")]
    [Parameter("ReadEndPoint", typeof(int), "the DECIMAL number used for the read endpoint", "129")]
    [Parameter("WriteEndPoint", typeof(int), "the DECIMAL number used for the write endpoint", "1")]
    [Parameter("timeout", typeof(int), "Read/Write timeout in milliseconds")]
    /*
     * Using LibUsbDotNet http://sourceforge.net/projects/libusbdotnet/
     * 
     */
    public class UsbDataPublisher : Publisher
    {

        #region Constructor
        public UsbDataPublisher(Dictionary<string, Variant> args)
            : base(args)
        {

        } 
        #endregion

        #region Variable declarations 
        
        public UsbDevice MyUsbDevice;
        
        private static NLog.Logger logger = LogManager.GetCurrentClassLogger();
        protected override NLog.Logger Logger { get { return logger; } }
        private Stream _Stream;
        private string _Result;
        public int VendorID { get; set; }
        public int ProductID { get; set; }
        public int Configuration { get; set; }
        public int Interface { get; set; }
        public int ReadEndPoint { get; set; }
        public int WriteEndPoint { get; set; }
        public int timeout { get; set; }
        public long SelfLength = 0;

        private UsbDeviceFinder MyUsbFinder = null;
        private ErrorCode ec = ErrorCode.None;
        private UsbEndpointReader reader = null;
        private UsbEndpointWriter writer = null;
        #endregion Variable declarations
        
        #region Required get and set variables for Peach
        public override long Length
        {
            get
            {
                return this._Stream.Length;
            }
        }
        public override long Position
        {
            get
            {
                return this._Stream.Position;
            }
            set
            {
                this._Stream.Position = value;
            }
        }

        public override string Result
        {
            get
            {
                return _Result;
            }
            set
            {
                this._Result = value;
            }
        }
        #endregion Required get and set variables for Peach

        #region implemented overidden methods 

        protected override void OnStart()
        {
            try
            {
                MyUsbFinder = new UsbDeviceFinder(VendorID, ProductID);
                MyUsbDevice = UsbDevice.OpenUsbDevice(MyUsbFinder);
                //Try for a while to get the device, just in case it is slow or something, reduce dependency in "plugging it in" timing
                if (MyUsbDevice == null)
                    for (int i = 0; i < 20; i++)
                    {
                        Thread.Sleep(200);
                        MyUsbFinder = new UsbDeviceFinder(VendorID, ProductID);
                        MyUsbDevice = UsbDevice.OpenUsbDevice(MyUsbFinder);
                        if (MyUsbDevice != null) break;
                    }
                if (MyUsbDevice == null)
                    //No point to continue if the device is not found after we tried so hard to find it
                    throw new PeachException(string.Format("USB Publisher: Device Not Found. Vendor: {0} Product: {1} Interface: {2}",
                        VendorID, ProductID, Interface));

                IUsbDevice wholeUsbDevice = MyUsbDevice as IUsbDevice;
                if (!ReferenceEquals(wholeUsbDevice, null))
                {
                    wholeUsbDevice.SetConfiguration(Convert.ToByte(Configuration));
                    wholeUsbDevice.ClaimInterface(Interface);
                }

                #region ReadEndPoint switch
                switch (ReadEndPoint)
                {

                    case 129: //0x81
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep01); //Ep1 is Ep 81 IRL
                        break;
                    case 130: //0x82
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep02); //Ep1 is Ep 81 IRL
                        break;
                    case 131: //0x83
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep03); //Ep1 is Ep 81 IRL
                        break;
                    case 132: //0x84
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep04); //Ep1 is Ep 81 IRL
                        break;
                    case 133: //0x85
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep05); //Ep1 is Ep 81 IRL
                        break;
                    case 134: //0x86
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep06); //Ep1 is Ep 81 IRL
                        break;
                    case 135: //0x87
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep07); //Ep1 is Ep 81 IRL
                        break;
                    case 136: //0x88
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep08); //Ep1 is Ep 81 IRL
                        break;
                    case 137: //0x89
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep09); //Ep1 is Ep 81 IRL
                        break;
                    case 138: //0x8A
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep10); //Ep1 is Ep 81 IRL
                        break;
                    case 139: //0x8B
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep11); //Ep1 is Ep 81 IRL
                        break;
                    case 140: //0x8C
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep12); //Ep1 is Ep 81 IRL
                        break;
                    case 141: //0x8D
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep13); //Ep1 is Ep 81 IRL
                        break;
                    case 142: //0x8E
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep14); //Ep1 is Ep 81 IRL
                        break;
                    case 143: //0x8F
                        reader = MyUsbDevice.OpenEndpointReader(ReadEndpointID.Ep15); //Ep1 is Ep 81 IRL
                        break;
                    default:
                        throw new PeachException(string.Format("USB Publisher: Read EndPoint definition init problem. Vendor: {0} Product: {1} Interface: {2}",
                            VendorID, ProductID, Interface));
                }
                #endregion

                #region WriteEndPoint switch

                switch (WriteEndPoint)
                {

                    case 1:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep01); //Ep1 is Ep 0x01 IRL
                        break;
                    case 2:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep02); //Ep2 is Ep 0x02 IRL
                        break;
                    case 3:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep03); //Ep3 is Ep 0x03 IRL
                        break;
                    case 4:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep04); //Ep4 is Ep 0x04 IRL
                        break;
                    case 5:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep05); //Ep5 is Ep 0x05 IRL
                        break;
                    case 6:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep06); //Ep6 is Ep 0x06 IRL
                        break;
                    case 7:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep07); //Ep7 is Ep 0x07 IRL
                        break;
                    case 8:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep08); //Ep8 is Ep 0x08 IRL
                        break;
                    case 9:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep09); //Ep9 is Ep 0x09 IRL
                        break;
                    case 10:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep10); //Ep10 is Ep 0x0A IRL
                        break;
                    case 11:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep11); //Ep11 is Ep 0x0B IRL
                        break;
                    case 12:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep12); //Ep12 is Ep 0x0C IRL
                        break;
                    case 13:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep13); //Ep13 is Ep 0x0D IRL
                        break;
                    case 14:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep14); //Ep14 is Ep 0x0E IRL
                        break;
                    case 15:
                        writer = MyUsbDevice.OpenEndpointWriter(WriteEndpointID.Ep15); //Ep15 is Ep 0x0F IRL
                        break;
                    default:
                        throw new PeachException("Write EndPoint deficnition init problem");
                }
                #endregion
                //base.OnStart();
            }
            catch (Exception ex)
            {
                throw new PeachException(string.Format("USB Publisher: Device Not Found."));
            }
        }

        protected override void OnStop()
        {
            if (MyUsbDevice != null)
            {
                if (MyUsbDevice.IsOpen)
                {
                    IUsbDevice wholeUsbDevice = MyUsbDevice as IUsbDevice;
                    if (!ReferenceEquals(wholeUsbDevice, null))
                    {
                        wholeUsbDevice.ReleaseInterface(Interface);
                    }

                    MyUsbDevice.Close();
                }
                MyUsbDevice = null;
                UsbDevice.Exit();
            }
            base.OnStop();

        }

        protected override void OnInput()
        {
            try
            {
                ErrorCode ecRead = ErrorCode.None;
                byte[] readBuffer = new byte[512];
                while (ecRead == ErrorCode.None)
                {
                    int bytesRead;
                    ecRead = reader.Read(readBuffer, 100, out bytesRead);
                    if (bytesRead != 0) //Console.WriteLine("Nothing Read, recieved 0 Bytes from read operation");
                    {
                        logger.Debug("RECIEVED " + bytesRead + " BYTES FROM USB!");
                        logger.Debug("HEX - USB Publisher DEBUG  --> " + BitConverter.ToString(readBuffer, 0, bytesRead).Replace("-", " "));
                        logger.Debug("ASCII - USB Publisher DEBUG  --> " + System.Text.Encoding.Default.GetString(readBuffer, 0, bytesRead));
                    }
                    
                    
                    this.Result = System.Text.Encoding.Default.GetString(readBuffer, 0, bytesRead);
                    this._Stream = new MemoryStream(readBuffer, 0, bytesRead, true, true);
                    this._Stream.Position = 0;
                    this.Position = 0;
                }
            }
            catch (Exception ex)
            {
                logger.Warn((ec != ErrorCode.None ? ec + ":" : String.Empty) + ex.Message);
            }
        }


        protected override void OnOutput(byte[] buffer, int offset, int count)
        {
            try
            {
                ErrorCode ecWrite;
                //We need to send count bytes from byffer starting at offset

                byte[] bytesToSend = new byte[count];
                for (int i = 0; i < count; i++)
                {
                    bytesToSend[i] = buffer[offset + i];
                }
                //data.Read(bytesToSend, 0, Convert.ToInt32(data.Length));
                int bytesWritten;
                ecWrite = writer.Write(bytesToSend, 100, out bytesWritten);
                if (ecWrite != ErrorCode.None) throw new Exception(UsbDevice.LastErrorString);
            }
            catch (Exception ex)
            {
                logger.Warn((ec != ErrorCode.None ? ec + ":" : String.Empty) + ex.Message);
            }
        }

    
        #endregion implemented overidden methods

        #region "Must have" overridden methods for the publisher

        public override void Flush()
        {
            this._Stream.Flush();
        }

        public override void SetLength(long value)
        {
            this._Stream.SetLength(value);
        }

        public override bool CanRead
        {
            get
            {
                return this._Stream.CanRead;
            }
        }

        public override bool CanWrite
        {
            get
            {
                return this._Stream.CanWrite;
            }
        }

        public override bool CanSeek
        {
            get
            {
                return this._Stream.CanSeek;
            }
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            this._Stream.Write(buffer, offset, count);
        }

        public override int ReadByte()
        {
            return this._Stream.ReadByte();
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            return this._Stream.Read(buffer, offset, count);
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            return this._Stream.Seek(offset, origin);
        }

        #endregion "Must have" overridden methods for the publisher

    }
}

// end
